#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

r"""
 _   _ ____________          _
| | | || ___ \ ___ \        (_)
| |_| || |_/ / |_/ / __ ___  _
|  _  ||  __/|  __/ '__/ _ \| |
| | | || |   | |  | | | (_) | |
\_| |_/\_|   \_|  |_|  \___/| |
                           _/ |
                          |__/
===============================
HealPix Projection
"""

__version__ = '0.8.1'

from .wcs_helper import *
from .hp_helper import *
from .cutsky import *
from .visu import *
